#include "gridstorefunctions.h"
#include "gridstore.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

// To install gsl library onto CentOS Linux 
// Use
// yum install epel-release
// yum install gsl-devel


//SIGMA represents a common standard deviation of temperature for drinking water

// Default GridDB Connection Settings 

// Serial Numbers at below URL
// https://www.randomcodegenerator.com/en/generate-serial-numbers






// Create a random but normally distributed random number based on a set mean (average) and set standard deviation (sigma)
double gaussian_number(double mean, double sigma,unsigned int seed){
	
	// Set up gsl environment
	gsl_rng *r; 
	const gsl_rng_type* range_type; 
	gsl_rng_env_setup(); // Setup gsl environment for creating random numbers

	// Set up gsl range and allocate resources
	range_type = gsl_rng_default; //Use default gsl range 
	r = gsl_rng_alloc(range_type);
	gsl_rng_set(r,seed); // Have the range set to time so each random number generated is always unique
	


	double generated_measurement = mean + gsl_ran_gaussian(r,sigma); // Offset by mean so the distribution has the desired average
	gsl_rng_free(r); // Clean up and free up allocated resources for gsl random number generation
	return generated_measurement;
}






// Bernoulli trials return 0 or 1 
// SUCCESS is denoted as 1 
// FAILURE is denoted as 0
// Method determines whether an espresso shot succeeds or fails based on the machine's failure rate
unsigned int determineEspressoSuccess(double failureProbability,unsigned int seed){
	gsl_rng *r; // Set up gsl
	const gsl_rng_type* range_type;
	gsl_rng_env_setup(); // Setup gsl environment for creating random numbers


	// Set up gsl range and allocate resources
	range_type = gsl_rng_default; // Use defaul gsl range
	r = gsl_rng_alloc(range_type); 
	gsl_rng_set(r, seed);// Have the range set to time so random numbers generated are always unique
	


	unsigned int espressoResult = gsl_ran_bernoulli(r,1.0 - failureProbability); // Obtain bernoulli result. Success probability is used in gsl. So we do 1-p to use it with failure rate
	gsl_rng_free(r); // Clean up and free resources allocated to random number generator
	

	return espressoResult;
}


int adjustedShotsperHour(int averageShots, double standard_deviation,unsigned int seed){
	double shots = gaussian_number(averageShots,standard_deviation,seed); // Return a number of shots that the coffee machine will output per hour
	// based on the machine's average hourly shot rate and deviation so far
	return (int)(shots); 
}




int getSleepTime(int shotCount, int shotRate, int secondsElapsed){


	int quotient = ((shotCount + 1)) / (shotRate * SECONDS_IN_HR);

	int sleepTime = quotient - secondsElapsed;

	if(sleepTime < 0)
		sleepTime = sleepTime * -1;
	else if(sleepTime == 0)
		sleepTime = 15;

	return sleepTime;
}




// Determine the amount of time an espresso machine must wait before between espresso shots
int calculateSecondsPerShot(int averageShots, double standard_deviation,unsigned int seed){
	int adjustShotRate = adjustedShotsperHour(averageShots, standard_deviation, seed); //Determine the amount of shots that will be made in the current hour
	
	double quotient = (double) (adjustShotRate / 3600.0); // Adjust by the number of seconds in an hour
	
	int seconds = (int) (1 / quotient); // Use integer reciprocal as answer
	
	if (seconds <= 0)
		seconds = seconds * -1;

	return seconds; 
}








// Create a connection to GriDB depending on GridDB settings and cluster properties
GSGridStore* makeGridStore(char* host, char* port, char* clusterName, char* username, char* password){
	GSGridStore* gridstore;


	

	//Set GridDB Cluster Properties
	const GSPropertyEntry properties[] = {
		{"notificationAddress",host},
		{"notificationPort",port},
		{"clusterName",clusterName},
		{"user",username},
		{"password",password}
	};
	size_t propCount = sizeof(properties) / sizeof(*properties);
	


	gsGetGridStore(gsGetDefaultFactory(),properties,propCount,&gridstore); // Connect to GridDB cluster
	return gridstore;
}




//Create and register a coffee chain for storing records of coffee stores
GSCollection* makeCoffeeStoreCollection(GSGridStore* gridstore, char* chainName){
	// Initialize Container and Dynamic Column Schema
	GSCollection* coffeeChain;
	GSColumnInfo columnInfo = GS_COLUMN_INFO_INITIALIZER;
	GSColumnInfo columnInfoList[4];
	GSContainerInfo containerInfo = GS_CONTAINER_INFO_INITIALIZER;
	


	// Create Column Schema
	columnInfo.name = "store_id"; // Primary key of coffee stores is a string column known as 'store_id'
	columnInfo.type = GS_TYPE_STRING;
	columnInfoList[0] = columnInfo;
	

	// Address Column
	columnInfo.name = "address"; // All stores have an 'address' location
	columnInfo.type = GS_TYPE_STRING;
	columnInfoList[1] = columnInfo;
	

	// Latitude Column
	columnInfo.name = "latitude"; // Stores also can be identified geographically (latitude and longitude)
	columnInfo.type = GS_TYPE_FLOAT;
	columnInfoList[2] = columnInfo;
	

	// Longitude Column
	columnInfo.name = "longitude";
	columnInfo.type = GS_TYPE_FLOAT;
	columnInfoList[3] = columnInfo;
	

	// Create Container Schema and Info Properties
	containerInfo.type = GS_CONTAINER_COLLECTION;
	containerInfo.name = chainName; //Name of container will be the name of the franchise or coffee chain
	containerInfo.columnCount = sizeof(columnInfoList) / sizeof(*columnInfoList); // Give container 4 columns
	containerInfo.columnInfoList = columnInfoList; // Create container's schema
	containerInfo.rowKeyAssigned = GS_TRUE; // Container will have a row-key (store_id)
	

	// Create Container and set Index
	gsPutCollectionGeneral(gridstore,NULL,&containerInfo,GS_FALSE,&coffeeChain);  // Insert container into GridDB
	gsCreateIndex(coffeeChain,"store_id",GS_INDEX_FLAG_DEFAULT); // Create a store_id index 
	

	return coffeeChain;
}





// All coffee stores can have 1 or multiple 'coffee machines' thus a container will keep record
// of all coffee machines stored within a coffee store.
GSCollection* makeMachineCollection(GSGridStore* gridstore, char* containerName){
	// Initialize container and Column Schema holders
	GSCollection* machineCollection;
	GSColumnInfo columnInfo = GS_COLUMN_INFO_INITIALIZER;
	GSColumnInfo columnInfoList[5];
	GSContainerInfo containerInfo = GS_CONTAINER_INFO_INITIALIZER;



	// Set up primary key column
	columnInfo.name = "serial_number"; // All coffee machines are identified by a unique serial number within a store
	columnInfo.type = GS_TYPE_STRING;
	columnInfoList[0] = columnInfo;



	// Set up model number column
	columnInfo.name = "model_number"; // All coffee machines also have model numbers depending on the type of device they are or their manufacturer
	columnInfo.type = GS_TYPE_STRING;
	columnInfoList[1] = columnInfo;


	// Set up installation date column schema
	columnInfo.name = "installation_date"; // Coffee Machines when first registered will be given a unique install date
	columnInfo.type = GS_TYPE_TIMESTAMP;
	columnInfoList[2] = columnInfo;


	// Set up Cycle Count Column Schema
	columnInfo.name = "cycle_since_new"; // Keeps track of how many espresso shots have been made since the coffee machine was installed
	columnInfo.type = GS_TYPE_INTEGER;
	columnInfoList[3] = columnInfo;


	//Set up Clean Cycle Count Column Schema
	columnInfo.name = "cycle_since_clean"; // Keeps track of how many espresso shots have been made since the coffee machine's last clean cycle.
	columnInfo.type = GS_TYPE_INTEGER;
	columnInfoList[4] = columnInfo;

	// Set Container Info Properties
	containerInfo.type = GS_CONTAINER_COLLECTION; // Make container type COLLECTION
	containerInfo.name = containerName; // Name the container by store_id of the Coffee store
	containerInfo.columnCount = sizeof(columnInfoList) / sizeof(*columnInfoList); //Make space for 5 columns
	containerInfo.columnInfoList = columnInfoList; // Create container schema
	containerInfo.rowKeyAssigned = GS_TRUE;  // Make the row-key the coffee machine's serial number
	



	// Create and Insert Container into GridDB and setup its Index
	gsPutCollectionGeneral(gridstore,NULL,&containerInfo,GS_FALSE,&machineCollection); // Insert collection into GridDB
	if(machineCollection != NULL)
		printf("Coffee Store with store ID: %s has been inserted into GridDB\n",containerName);
	gsCreateIndex(machineCollection,"serial_number",GS_INDEX_FLAG_DEFAULT);
	return machineCollection;
}





// Coffee or Espresso Machines make shots every few seconds / minutes 
// Each shot is espressed as a timestamped record that is sent to GridDB
GSTimeSeries* makeEspressoMachine(GSGridStore* gridstore, char* serialNumber){
	
	// Initialize Timeseries and column schema holders
	GSTimeSeries* espressoMachine;
	GSColumnInfo columnInfo = GS_COLUMN_INFO_INITIALIZER;
	GSColumnInfo columnInfoList[4];
	GSContainerInfo containerInfo = GS_CONTAINER_INFO_INITIALIZER;



	// Setup Timestamp Column Schema
	columnInfo.name = "timestamp"; // Timestamp column
	columnInfo.type = GS_TYPE_TIMESTAMP;
	columnInfoList[0] = columnInfo;


	// Setup Event Type Column Schema
	columnInfo.name = "event_type"; // An espresso shot can be classified three ways: as a SUCCESS, as a FAILURE, or in a CLEAN cycle
	columnInfo.type = GS_TYPE_INTEGER;
	columnInfoList[1] = columnInfo;


	// Setup Temperature Column Schema 
	columnInfo.name = "temperature"; // Water temperature of the current espresso shot
	columnInfo.type = GS_TYPE_FLOAT;
	columnInfoList[2] = columnInfo;


	// Set up Pressure Column Schema
	columnInfo.name = "pressure"; // Water pressure of the current espresso shot
	columnInfo.type = GS_TYPE_FLOAT;
	columnInfoList[3] = columnInfo;


	// Setup Container Info Values
	containerInfo.type = GS_CONTAINER_TIME_SERIES; // Container will be a timeseries container
	containerInfo.name = serialNumber;
	containerInfo.columnCount = sizeof(columnInfoList) / sizeof(*columnInfoList); // Make space for 4 columns
	containerInfo.columnInfoList = columnInfoList; // Create container schema
	containerInfo.rowKeyAssigned = GS_TRUE; // Give the container a row-key of timestamp
	

	// Insert Timeseries container into GridDB 
	gsPutTimeSeriesGeneral(gridstore,NULL,&containerInfo, GS_FALSE,&espressoMachine); // Insert timeseries into GridDB
	if(espressoMachine != NULL)
		printf("Coffee Machine with serial number: %s has been inserted into GridDB\n",serialNumber);
	return espressoMachine;
}






void insertCoffeeStore(GSCollection* coffeeChain, char* store_id, char* address, float latitude, float longitude){
	GSRow* storeEntry; // A row is a coffee store in a coffee chain container
	gsCreateRowByContainer(coffeeChain,&storeEntry);
	


	gsSetRowFieldByString(storeEntry,0,store_id); // Set store's id
	gsSetRowFieldByString(storeEntry,1,address); // Set store's addresss
	gsSetRowFieldByFloat(storeEntry,2,latitude); // Set store's latitude
	gsSetRowFieldByFloat(storeEntry,3,longitude);// Set store's geographic longitude
	


	gsPutRow(coffeeChain,NULL,storeEntry,NULL); // Insert coffee store into coffee chain
	gsCloseRow(&storeEntry); // Clean up
}






void insertCoffeeMachine(GSCollection* coffeeStore, char* serial_number, char* model_number, GSTimestamp installation_date, int cycles_new, int cycles_clean){
	GSRow* coffeeMachine; // A row has information on a coffee machine in a coffee store container
	gsCreateRowByContainer(coffeeStore, &coffeeMachine); // Get row with the coffee store row schema
	


	gsSetRowFieldByString(coffeeMachine,0,serial_number); // Set machine's serial number
	gsSetRowFieldByString(coffeeMachine,1,model_number); // Set machine's model number
	gsSetRowFieldByTimestamp(coffeeMachine,2,installation_date); // Set machine's install date
	gsSetRowFieldByInteger(coffeeMachine,3,cycles_new); // Set the amount of cycles a machine has since new
	gsSetRowFieldByInteger(coffeeMachine,4,cycles_clean); // Set the amount of cycles a machine has had since its last clean cycle
	

	gsPutRow(coffeeStore,NULL,coffeeMachine,NULL); // Insert coffee machine into coffee store
	gsCloseRow(&coffeeMachine);
}








// Return the string-version of an event-state integer code
char* determineEventType(int eventState){
	
	switch(eventState){
		case SUCCESS:
			return "SUCCESS";
		



		case FAILURE: 
			return "FAILURE";




		case CLEAN:
			return "CLEAN";
	}
	return "";
}




// Updates the cycle counts for coffee machine record in the coffee store container
// the coffee machine belongs to.
void updateMachineCycles(GSCollection* coffeeStore,char* serialNumber,bool resetCleanCycle){
	GSRow* machineRecord; // Instantiate Coffee Machine Record row
	int new_cycles, clean_cycles;
	


	gsCreateRowByContainer(coffeeStore,&machineRecord); // Get Row Schema
	gsGetRowByString(coffeeStore,serialNumber,machineRecord,GS_TRUE,NULL); // Get coffee machine row that the coffee machine belongs to
	


	gsGetRowFieldAsInteger(machineRecord,3,&new_cycles); // Get number of cycles since coffee machine was installed
	gsGetRowFieldAsInteger(machineRecord,4,&clean_cycles); // Get number of cycles since coffee machine was last cleaned
	


	gsSetRowFieldByInteger(machineRecord,3,++new_cycles); // Update new cycles 
	
	if(resetCleanCycle) // If the last cycle was a clean cycle
		gsSetRowFieldByInteger(machineRecord,4,0); // Reset clean cycle count to zero
	else
		gsSetRowFieldByInteger(machineRecord,4,++clean_cycles); // Update clean cycles if last cycle was not a clean cycle
	
	
	gsPutRow(coffeeStore,NULL,machineRecord,NULL); // Update row
	
	gsCloseRow(&machineRecord); // Clean up
}




// Make an espresso shot record and insert into GridDB based on random factors and espresso machine setttings like failure_rate and if it needs to be 
// on a clean cycle
void makeEspressoRecord(GSTimestamp timestamp,GSTimeSeries* espressoMachine,GSCollection* coffeeStore, char* serialNumber,double average_temperature, double average_pressure, double failure_rate,unsigned int seed,bool isCleanCycle){
	float recorded_temperature = (float) gaussian_number(average_temperature,SIGMA,seed); // Generate a random temperature that is within a set average and deviation
	float recorded_pressure = (float) gaussian_number(average_pressure,PRESSURE_SIGMA,seed); // Generate a random pressure that is within a set average and deviation
	int eventState; // Determine which cycle the current cycle
	


	if(isCleanCycle){ 
		eventState = CLEAN; // Set the cycle to clean if it is time  to be on a clean cycle
	} else {
		eventState = determineEspressoSuccess(failure_rate,seed); // Determine whether or not the espresso shot was a success based on the machine's failure rate
	}
	

	//GSTimestamp timestamp = gsCurrentTime(); // Make time stamp record
	GSChar buffer[GS_TIME_STRING_SIZE_MAX]; // Formatted timestamp for display
	gsFormatTime(timestamp,buffer,sizeof(buffer)); 
	

	printf("Inserting record at %s, with state %s and measurements of a temperature of %f and a pressure of %f into Timeseries %s\n",buffer,determineEventType(eventState),recorded_temperature,recorded_pressure, serialNumber);
	GSRow* espressoShotRecord; // Create Row and obtain row schema
	gsCreateRowByContainer(espressoMachine,&espressoShotRecord);// check if gsPut works with TimeSeries; Create a row with a coffee machine schema
	


	// Set timestamp record column values
	gsSetRowFieldByTimestamp(espressoShotRecord,0,timestamp); // Set the timestamp of the record
	gsSetRowFieldByInteger(espressoShotRecord,1,eventState); // Set the eventType
	gsSetRowFieldByFloat(espressoShotRecord,2,recorded_temperature); // Set the temperature
	gsSetRowFieldByFloat(espressoShotRecord,3,recorded_pressure); // Set the pressure
	

	gsPutRow(espressoMachine,NULL,espressoShotRecord,NULL); // Insert row into TimeSeries
	



	gsCloseRow(&espressoShotRecord); // Clean up and update the coffee machine record in the coffee store container the coffee machine belongs to
	updateMachineCycles(coffeeStore,serialNumber,isCleanCycle);
}