#!/usr/bin/python
import griddb_python_client
from datetime import date, timedelta,datetime
import numpy as np
import time
import sys, traceback


griddb = griddb_python_client

SUCCESS = 1
CLEAN = 2
FAILURE = 0

MILLISECS_IN_HR = 1000 * 60 * 60 

Timestamp_current = griddb.Timestamp_current

# List to include 
# fetchGridstore, get_store_ids, get_serial_nums, get_past_shot_count, get_time, get_avg_shot_per_hour, get_recent_readings, 
# get_machine_clean_rate, fetch_avg_readings, fetch_machine_cycles, get_recent_event, get_instantatneous_rate, fetch_timestamps_betw



def fetchGridStore(host, port, cluster, user, password):
	"""Create a connection to a GridDB cluster"""
	try:
		factory = griddb.StoreFactory.get_default()
		gridstore = factory.get_store({
			"notificationAddress":host,
			"notificationPort": port,
			"clusterName":cluster,
			"user":user,
			"password":password
			})
		return gridstore
	except:
		print("Griddb connection failed!")
		traceback.print_stack()




def get_store_ids(coffee_chain):
	"""Obtain all store_ids of the coffee stores in a coffee chain"""
	store_ids = [] ## Coffee Stores
	try:
		tql = "select *"
		query = coffee_chain.query(tql)
		rowSet = query.fetch(False) ## Obtain and retrieve all coffee stores in the coffee chain
		row = coffee_chain.create_row()


		while rowSet.has_next():
			rowSet.get_next(row)  ### Get coffee store record
			store_id = row.get_field_as_string(0)  ## Retrieve store id
			store_ids.append(store_id) ## Obtain and add the coffee store's store id to the collection of stores available to view
	


	except:
		print("Either GridDB failed to connect or the Chain Name you provided is invalid")
		traceback.print_stack()
		sys.exit(1) ## If the coffee chain is invalid or nonexistent then there are no plots or graphs to display
	return store_ids




def get_serial_nums(gridstore,store_id):
	"""Get a list of all serial numbers for all coffee machines in a coffee store with the given store id"""
	serial_nums = []
	

	try:
		coffee_store = gridstore.get_container(store_id) ### Fetch Coffee store container
		query = coffee_store.query("select *") ### Query and fetch all coffee machine rows in the coffee store container
		rowSet = query.fetch(False)
		row = coffee_store.create_row()


		while rowSet.has_next():
			rowSet.get_next(row)  ### Get coffee machine info
			serial_num = row.get_field_as_string(0)
			serial_nums.append(serial_num)  ### Obtain coffee machine's serial number and add it to the serial number collection
	


	except:
		print("Failed to fetch coffee machine serial numbers for store: " + store_id)
		traceback.print_stack()
	
	return serial_nums




def get_past_shot_count(coffee_machine,event_type,setting,offset,end):
	"""Obtains the amount of shots a coffee machine made x hours since the current time"""
	tql = "select COUNT(*) where event_type={0} AND timestamp >= TIMESTAMPADD({1},NOW(),{2}) AND timestamp <= TIMESTAMPADD({1},NOW(),{3})".format(event_type,setting,-1 * offset, -1 * end) ### Obtain successful shots from between (offset) hours ago and 
																											### (offset -1) hours ago
	shot_count = 0
	try:
		query = coffee_machine.query(tql) ## Create query
		rowSet = query.fetch(False) ## Fetch query



		aggregation = rowSet.get_next_aggregation() ## Obtain aggregation
		shot_count = aggregation.get_long()  ### Obtain numerical result from aggregation
	except:
		print("Failed to get shot count from " + str(offset) + " hours ago.")
		traceback.print_stack()

	return shot_count




def get_time(coffee_machine,setting):
	"""Get either the earliest or latest timestamp from a coffee machine Timeseries"""
	time = 0
	

	try:
		tql = "select " ## Get either latest or earliest timestamp for a coffee machine
		if setting == "MAX":
			tql += "max(timestamp)" ## Get latest timestamp
		elif setting == "MIN":
			tql += "min(timestamp)" ## Get earliest timestamp


		query = coffee_machine.query(tql) 
		rowSet = query.fetch(False) ## Create and fetch the time query


		aggregation = rowSet.get_next_aggregation() ## Get time-type aggregation result
		time = aggregation.get_timestamp() ## Obtain the timestamp from the aggreagtion Value


	except:
		print("Failed to retrieve timestamps, ensure the container has the correct row schema")
		traceback.print_stack()


	return time





def get_shots_since(coffee_machine,timestamp = griddb.Timestamp_current()):
	"""Get number of shots since a certain timestamp"""
	shots = 0
	try:
		tql = "select COUNT(*) where event_type=1 AND timestamp <= TO_TIMESTAMP_MS({0})".format(timestamp)
		query = coffee_machine.query(tql)
		rowSet = query.fetch(False)
		if rowSet.has_next():
			aggregation = rowSet.get_next_aggregation()
			shots = aggregation.get_long()
	except:
		print("Failed to retrieve shot count")
		traceback.print_stack()

	return shots



def fetch_old_averages(coffee_machine,timestamp = griddb.Timestamp_current()):
	"""Get an old shot per hour estimate for a past timestamp"""
	shot_count = get_shots_since(coffee_machine,timestamp) ## Get the amount of shots that were issued before that timestamp
	#hours = fetch_old_data(coffee_machine,timestamp) ### Find the amount of hours that elapsed since that timestamp and since the coffee machine was used
	
	earliest_time = get_time(coffee_machine,"MIN")
	hours = ( float(timestamp) - earliest_time) / (float(MILLISECS_IN_HR))
	if hours > 0:
		return shot_count / float(hours)
	return 0



def get_avg_shot_per_hour(coffee_machine):
	"""Obtain hourly shot rate for a coffee machine"""
	timestamp = griddb.Timestamp_current()
	shot_count = get_shots_since(coffee_machine,timestamp)
	#hours = fetch_time_sampling(coffee_machine)
	currentTime = griddb.Timestamp_current()
	earliest_time = get_time(coffee_machine,"MIN")
	hours = (float(currentTime) - earliest_time) / (float(MILLISECS_IN_HR))
	if hours > 0:
		return float(shot_count) / (hours)
	return 0





### Temperature and Pressure Queries into GridDB  ####
def get_recent_readings(gridstore,serial_num,timestamp = griddb.Timestamp_current()):
	"""Obtain the most recent temperature for a coffee machine """
	readings = {"temperature":0,"pressure":0}
	try:
		coffee_machine = gridstore.get_container(serial_num)  ### Obtain coffee machine container


		tql = "select * where timestamp <= TO_TIMESTAMP_MS({0}) order by timestamp desc limit 1".format(timestamp) ### Obtain most recent timestamp record from the coffee machine
		query = coffee_machine.query(tql) ### Create the query
		rowSet = query.fetch(False) ### Fetch the timestamp query


		row = coffee_machine.create_row()
		if rowSet.has_next():
			rowSet.get_next(row) ## Obtain the most recent timestamp record



			temperature = row.get_field_as_float(2) ## Retrieve the recent temperature
			pressure = row.get_field_as_float(3) ## Retrieve the recent pressure


			readings["temperature"] = temperature
			readings["pressure"] = pressure
		
	except:
		print("Either a container with name: " + serial_num + " does not exist or has an invalid row-schema")
		traceback.print_stack()


	return readings




def get_machines_readings(gridstore,store_id):
	"""Obtain the averages of the most recent temperatures and pressures from a coffee store"""
	coffee_store = gridstore.get_container(store_id) ## Get coffee store container with the store_id as its name
	readings = {"temperature":0 , "pressure": 0}
	try:
		get_all_tql = "select *"
		count = 0  ### Number of coffee machines
		avg_temperature = 0
		

		query = coffee_store.query(get_all_tql)
		rowSet = query.fetch(False)  ### Obtain all coffee machines from a coffee store container
		coffee_machine_row = coffee_store.create_row()
		

		while rowSet.has_next():
			count += 1
			rowSet.get_next(coffee_machine_row)  ### Obtain coffee machine information
			serial_num = coffee_machine_row.get_field_as_string(0)  ### Obtain coffee machine's serial number
			
			try:
				coffee_machine = gridstore.get_container(serial_num) ### Get the timeseries container with the serial_number as its container name
				

				tql = "select * order by timestamp desc limit 1" ## Get most recent timestamp record
				timestamp_query = coffee_machine.query(tql) ### Create the query
				record_set = timestamp_query.fetch(False) ## Fetch the query


				coffee_record = coffee_machine.create_row()
				record_set.get_next(coffee_record) ### Get timestamp record


				temperature = coffee_record.get_field_as_float(2) ### Obtain the temperature 
				pressure = coffee_record.get_field_as_float(3) ### Obtain the pressure
				


				readings["temperature"] += temperature ### Sum all the recent temperatures of every coffee machine
				readings["pressure"] += pressure  ### Sum all the recent pressures of every coffee machine
			


			except:
				print("Failed to connect to coffee machine: " + serial_number)

		for key in readings:
			readings[key] = readings[key] / float(count) ## Obtain average readings 


	except:
		print("The container for store " + store_id + " is either nonexistent or invalid")
		traceback.print_stack()


	return readings




def get_machine_clean_rate(coffee_machine):
	"""Get average time a coffee machine waits for a clean cycle"""
	clean_rate = 0
	clean_times = []
	

	try:
		earliest_time = get_time(coffee_machine,"MIN")  ### Get coffee machine's earliest timestamp 
		latest_time = get_time(coffee_machine,"MAX")  ### Get coffee machine's latest timestamp
		last_time = earliest_time  ### Start from the earliest and progress to the latest timestamp
		


		row = coffee_machine.create_row()   ### Query and fetch all records where there was a clean cycle
		query = coffee_machine.query("select * where event_type=2 order by timestamp asc")
		rowSet = query.fetch(False)

		while rowSet.has_next():
			rowSet.get_next(row)  ### Obtain timestamp record
			timestamp = row.get_field_as_timestamp(0)  ## Retrieve timestamp from row
			

			time_diff = timestamp - last_time  ### Measure time between clean cycles
			clean_times.append(time_diff)  ### Add time difference to the collection
			last_time = timestamp  ### Update the timestamp to compare for the next row

		if len(clean_times) > 0:
			clean_rate = np.mean(clean_times)  ### Average all differences to find average time between clean cycles 
	except:
		print("Query failed")
		traceback.print_stack()


	return clean_rate



def fetch_avg_readings(coffee_machine):
	"""Retrieve average temperature and average pressures for a coffee machine"""
	readings = {"temperature":0, "pressure":0}


	try:
		temperature_query = coffee_machine.query("select TIME_AVG(temperature)")  ## Query and fetch for the time-weighted average temperature
		temperature_rs = temperature_query.fetch(False)


		temperature_aggregation = temperature_rs.get_next_aggregation()  ## Obtain AggregationResult and retrieve numerical result
		temperature = temperature_aggregation.get_double()



		pressure_query = coffee_machine.query("select TIME_AVG(pressure)") ## Query and fetch for the time-weighted average pressure
		pressure_rs = pressure_query.fetch(False)



		pressure_aggregation = pressure_rs.get_next_aggregation()  ### Obtain AggregationResult and retrieve numerical result
		pressure = pressure_aggregation.get_double()


		readings["temperature"] = temperature  ## Update with averaged readings
		readings["pressure"] = pressure
	

	except:
		print("Failed To fetch average temperature and coffee machine readings")
		traceback.print_stack()


	return readings





def fetch_machine_cycles(coffee_store,serial_num):
	"""Retrieve the number of cycles and clean cycles for a coffee machine"""
	cycles = {"clean": 0, "new":0}
	

	try:
		tql = "select * where serial_number='" + serial_num + "'"  ### Find coffee machine record with the following serial number


		query = coffee_store.query(tql)  ### Query and fetch for that coffee machine
		rowSet = query.fetch(False)
		row = coffee_store.create_row()	


		if rowSet.has_next():
			rowSet.get_next(row)  ## Obtain coffee machine record
			

			cycles_new = row.get_field_as_integer(3) ### Obtain new cycle and clean cycle counts from the coffee machine row
			cycles_clean = row.get_field_as_integer(4)


			cycles["clean"]  = cycles_clean  ### Update with cycle counts
			cycles["new"] = cycles_new

	except:
		print("Failed to obtain row for coffee machine " + serial_num)
		traceback.print_stack()

	return cycles




def get_recent_event(coffee_machine,event_type):
	"""Get most recent timestamp record with a certain event_type from a coffee machine"""
	current_time = griddb.Timestamp_current()  ### Obtain current time
	recent_time = 0
	


	try:
		tql = "select * where event_type={0} order by timestamp desc limit 1".format(event_type) ### Request for most recent record with said event_type
		recent_time = get_time(coffee_machine,"MIN")  ## Get earliest record if there are no records in the coffee machine with event_type
	


		query = coffee_machine.query(tql)  ### Query and fetch for most recent record
		rowSet = query.fetch(False)
		row = coffee_machine.create_row()
	

		if rowSet.has_next():
			rowSet.get_next(row) ## Obtain record
	


			timestamp = row.get_field_as_timestamp(0)   ### Obtain timestamp from the row
			recent_time = current_time - timestamp  ## Obtain time difference to see how long it has been a coffee machine has said 'event_type'
	
	except:
		print("Error retrieving values from coffee machine")
		traceback.print_stack()

	return recent_time



def recent_state(gridstore,serial_numbers,event_type):
	"""Get most recent record of a certain event_type out of all the coffee machine in a coffee store"""	
	current_time = griddb.Timestamp_current()  ### Fetch the current time
	min_time_diff = current_time
	machine_diffs = []


	for serial_number in serial_numbers:
		coffee_machine = gridstore.get_container(serial_number) ### Obtain coffee machine Timeseries with the serial number as its name

		try:
			machine_time_diff = get_recent_event(coffee_machine,event_type)  ## Get time difference for the coffee machine
			machine_diffs.append(machine_time_diff)


		except:
			print("Failed to fetch data from coffee machine: "  + serial_number)
			tracback.print_stack()

	if len(machine_diffs) > 0:
		min_time_diff = min(machine_diffs)  ### Obtain the most recent (lowest) time difference out of all the coffee machine

	return min_time_diff




def get_instantaneous_rate(coffee_machine,timestamp = griddb.Timestamp_current(), maximum = 100):
	rate = 0
	
	try:
		past_timestamp = timestamp


		past_row = coffee_machine.create_row()

		pastql = "select * where event_type=1 AND timestamp < TO_TIMESTAMP_MS({0}) order by timestamp desc limit 1".format(timestamp)
		

		pastQuery = coffee_machine.query(pastql)
		pastRs = pastQuery.fetch(False)
		while pastRs.has_next():
			pastRs.get_next(past_row)
			past_timestamp = past_row.get_field_as_timestamp(0)


		delta = float(( timestamp - past_timestamp )) / ( 1000.0 )

		if delta > 0:
			rate = ( 1 / float(delta) ) * 3600



	except:
		print("Failed to fetch record at " + str(timestamp))
		traceback.print_stack()

	return rate


def fetch_timestamps_betw(gridstore,serial_number,start,end = griddb.Timestamp_current(), filtered = False, setting = SUCCESS):
	tql = "select * where timestamp >= TO_TIMESTAMP_MS({0}) and timestamp < TO_TIMESTAMP_MS({1}) order by timestamp desc".format(start,end)
	timestamps = []
	if filtered:
		tql = "select * where event_type={0} and timestamp >= TO_TIMESTAMP_MS({1}) and timestamp <= TO_TIMESTAMP_MS({2}) order by timestamp desc".format(setting,start,end)


	try:
		coffee_machine = gridstore.get_container(serial_number)
		query = coffee_machine.query(tql)

		rowSet = query.fetch(False)
		row = coffee_machine.create_row()
		while rowSet.has_next():
			rowSet.get_next(row)
			timestamp = row.get_field_as_timestamp(0)
			timestamps.append(timestamp)

	except:
		traceback.print_stack()
		print("Failed fetching records from {0} to {1}".format(start,end))

	return timestamps