import sys, traceback
import griddb_python_client
from datetime import date, timedelta, datetime
import time
import random
import math
import GridstoreMethods
from bokeh.io import show, curdoc
from bokeh.events import PanEnd,PanStart,Pan,LODStart,LODEnd, Press, Tap, MouseEnter
from bokeh.plotting import figure, output_file
from bokeh.layouts import widgetbox, column, layout, Spacer, row
from bokeh.models import ColumnDataSource, Div, Range1d, DatetimeTickFormatter, DataRange1d, Button, ResetTool, HelpTool, Legend, HoverTool, ZoomInTool, ZoomOutTool
from bokeh.core.property.containers import PropertyValueDict
from bokeh.models.widgets import Select, Panel, Tabs
from bokeh.client import push_session
from bokeh.driving import linear
from bokeh.models.callbacks import CustomJS


#### How to launch the data visualisation component from the command line ##### 

## Command Line Arguments (For now)
## 1. Coffee Chain Name
## 2. Preload [y or n] 
## 3. How far to prefetch from (in hours)

## Format
# $ python -m bokeh serve modulename.py --show --args $COMMAND_LINE_ARGS

## Example 
# $ python -m bokeh serve plotTimeseries.py --show --args oc_sample_2 y 24



LINE_COLORS = ["red","blue","green","purple","orange","yellow","black","#800000","#CD5C5C","#DC143C", "#8B0000", "#DB7093", "#FF1493", 
"#FF4500","#FFFF00", "#BDB76B","#FF00FF", "#663399", "#8B008B" ,"#4B0082", "#32CD32" ,"#006400" ,"#008B8B", "#4682B4" ,"#00BFFF"]
COLOR_COUNT = 25
### GridDB Default Connection Settings
HOST = "239.0.0.1"
PORT = "31999"
CLUSTER_NAME = "defaultCluster"
USER = "admin"
PASSWORD = "admin"


MILLISECS_IN_HOUR = 1000 * 60 * 60

DEFAULT_INTERVAL = 1000 * 60 * 60 * 24

#Default Plot Tools Provided Bokeh Library
TOOLS = "save,xpan,box_zoom,undo,redo,xwheel_zoom"

chain_name = "oc_sample_2"





#### Webpage Methods
def get_past_shot_rates(gridstore,serial_number):
	coffee_machine = gridstore.get_container(serial_number)
	x = []
	y = []
	serial_numbers = []
	tql = "select * where event_type=1 order by timestamp asc"
	query = coffee_machine.query(tql)
	rowSet = query.fetch(False)
	row = coffee_machine.create_row()
	while rowSet.has_next():
		rowSet.get_next(row)
		timestamp = row.get_field_as_timestamp(0)
		shotRate = GridstoreMethods.get_instantaneous_rate(coffee_machine,timestamp)
		x.append(datetime.fromtimestamp(int(timestamp) / 1000))
		y.append(shotRate)
		serial_numbers.append(serial_number)
	return x, y, serial_numbers

def get_past_temperatures(gridstore,serial_number):
	coffee_machine = gridstore.get_container(serial_number)
	x = []
	y = []
	serial_numbers = []
	tql = "select * order by timestamp asc"
	query = coffee_machine.query(tql)
	rowSet = query.fetch(False)
	row = coffee_machine.create_row()
	while rowSet.has_next():
		rowSet.get_next(row)
		timestamp = row.get_field_as_timestamp(0)
		temperature = row.get_field_as_float(2)
		x.append(datetime.fromtimestamp(int(timestamp) / 1000))
		y.append(temperature)
		serial_numbers.append(serial_number)
	return x , y,serial_numbers


def offline_shotrates(gridstore,store_id,serial_numbers):
	items = []
	timeseries = figure(name=store_id + " Toggled Plot",title = "Machines in Coffee Store " + store_id,x_axis_type = 'datetime',x_axis_label = "Current Time(Calendar)" , x_range=DataRange1d(follow="end",range_padding=0.0,follow_interval=DEFAULT_INTERVAL),
	y_range = Range1d(15,90), y_axis_label="Shots Per Hour",plot_width=1000,plot_height=600,tools=TOOLS,toolbar_location="right")
	for idx in range(len(serial_numbers)):
		x,y, machine= get_past_shot_rates(gridstore,serial_numbers[idx])
		#shot_circle = timeseries.circle("x","y",source = ColumnDataSource(data = dict(x = x, y = y, machine = machine)),line_color = LINE_COLORS[idx % COLOR_COUNT],fill_color = 'white',size= 3)
		shot_line = timeseries.line("x","y",source = ColumnDataSource(data = dict(x = x, y = y)), color=LINE_COLORS[idx % COLOR_COUNT])
		
		shot_line_source = shot_line.data_source ### Data source of timeseries line that can be updated and changed in callback functions
		items.append(("Coffee Machine: " + serial_numbers[idx],[shot_line]))

		## Add legend and its layout ##
	legend = Legend(items = items, location = (30,-30),label_width=200)
	timeseries.add_layout(legend,'right')


	xformatter = DatetimeTickFormatter(hours = ["%a %H:%M:%S"],minutes = ["%H:%M:%S"],seconds = ["%H:%M:%S"]) ### Have datetimes appear as formatted calendar dates and times
	timeseries.xaxis.formatter = xformatter
	timeseries.xaxis.major_label_orientation = "vertical"   ### Rotate x ticks so they can be easier to read
	
	## Set plot styling settings ##
	timeseries.min_border_bottom = 100
	timeseries.title.align = 'left'
	

	
	hover = HoverTool(tooltips = [("Time: ", "@x{%F}"),("Shot Rate: ", "$y shots per hr")],formatters=dict(x='datetime'))
	## Add extra tools to plot toolbar ##
	timeseries.add_tools(hover)
	timeseries.add_tools(ZoomInTool(dimensions = 'width', factor = 0.25))
	timeseries.add_tools(ZoomOutTool(dimensions = 'width', factor = 0.25))
	timeseries.add_tools(ResetTool(reset_size=True))
	timeseries.add_tools(HelpTool(help_tooltip="Learn more about GridDB at their official website",redirect="https://griddb.net/en"))
	return timeseries


def offline_temperatures(gridstore,store_id,serial_numbers):
	timeseries = figure(name=store_id + " Temperature Timeseries",title = "Machines in Coffee Store " + store_id, x_axis_type="datetime",x_axis_label="Current Time(Calendar)",y_range=Range1d(160,200),
	x_range=DataRange1d(follow="end",range_padding=0.0,follow_interval = DEFAULT_INTERVAL),y_axis_label="Current Temperature (Fahrenheit)", plot_width = 1000,plot_height=800,tools=TOOLS,toolbar_location='right')
	x_axes = []
	items = []
	y_axes = []
	for idx in range(len(serial_numbers)):
		x,y, machine = get_past_temperatures(gridstore,serial_numbers[idx])
		#shotcircle = timeseries.circle("x","y",source = ColumnDataSource(dict(x = x, y = y)), line_color = LINE_COLORS[idx % COLOR_COUNT],fill_color = 'white',size=3)
		shot_line = timeseries.line("x","y",source = ColumnDataSource(dict(x = x, y = y)),color = LINE_COLORS[idx % COLOR_COUNT])
		shot_line_source = shot_line.data_source
		items.append(("Coffee Machine: " + serial_numbers[idx],[shot_line]))

	## Add legend and its layout to plot ##
	legend = Legend(items = items, location = (30,-30),label_width = 200)
	timeseries.add_layout(legend,'right')

	xformatter = DatetimeTickFormatter(hours = ["%a %H:%M:%S"],minutes = ["%H:%M:%S"],seconds = ["%H:%M:%S"]) ### Have datetimes appear as formatted calendar dates and times
	timeseries.xaxis.formatter = xformatter
	timeseries.xaxis.major_label_orientation = 'vertical'	
	
	## Add plot style properties ##
	timeseries.min_border_bottom = 100
	timeseries.title.align = 'left'

	#timeseries.x_range.on_change('start',findPreviousTemperatures) ## Add pan callback event handler
	

	hover = HoverTool(tooltips = [("Date: ","@x{%F}"),("Shot Rate: ","$y Shots Per Hr.")],formatters=dict(x='datetime'))
	timeseries.add_tools(hover)
	## Add extra tools to plot toolbar ## 
	timeseries.add_tools(ZoomInTool(dimensions = 'width', factor = 0.25))
	timeseries.add_tools(ZoomOutTool(dimensions = 'width', factor = 0.25))
	timeseries.add_tools(ResetTool(reset_size=True))
	timeseries.add_tools(HelpTool(help_tooltip="Learn more about GridDB at their Official Website", redirect="https://griddb.net/en/"))
	return timeseries

def makeTimeseriesTabs(gridstore,store_ids):
	panels = []
	for store_id in store_ids:
		serial_numbers = filter(lambda serial_num: len(serial_num) > 0, GridstoreMethods.get_serial_nums(gridstore,store_id))
		shot_plot = offline_shotrates(gridstore,store_id,serial_numbers)
		temperature_plot = offline_temperatures(gridstore,store_id,serial_numbers)
		layout = column(shot_plot,temperature_plot,name = store_id)
		panel = Panel(child = layout, title = store_id)
		panels.append(panel)
	return panels




## Connect to Main Coffee Chain
gridstore = GridstoreMethods.fetchGridStore(HOST,PORT,CLUSTER_NAME,USER,PASSWORD)
coffee_chain = gridstore.get_container(chain_name)

store_ids = filter(lambda store: len(store) > 0, GridstoreMethods.get_store_ids(coffee_chain))  ### Get all coffee stores in the coffee chain

div = Div(text = "<h1 style='text-align:center;'>Plots for Coffee Chain: " + chain_name + "</h1>",height = 100, width = 1000)


panels = makeTimeseriesTabs(gridstore,store_ids)

tabs = Tabs(tabs = panels)

layout = column(div,tabs,name = "mainLayout", responsive = True)
show(layout)


output_file("timeseries_report.html")